//
//  MIDI.m
//  CocoaMIDI
//
//  Created by Paul Dorman on Sun Feb 20 2005.
//  pdorman@umich.edu
//  Public Domain. Feel free to use this however you want. :)
//

#import "MIDI.h"

NSMutableDictionary *chords;	// load chords from file and store them here

@implementation MIDI

+ (void) initialize{
	int i;

	NSString *chordFile = [[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent] stringByAppendingString:@"/Chords.txt"];
	NSString *chordDefs = [NSString stringWithContentsOfFile:chordFile];
	NSArray *eachChord = [chordDefs componentsSeparatedByString:@"\n"];	// each line
	chords = [[NSMutableDictionary alloc] initWithCapacity:7];
	for( i = 0; i < [eachChord count]; i++ ){
		NSArray *temp, *keys;

		if( [[eachChord objectAtIndex:i] isEqualToString:@""] )
			continue;	// don't crash if Chords.txt has extra returns

		temp = [[eachChord objectAtIndex:i] componentsSeparatedByString:@"\t"];
		keys = [[temp objectAtIndex:1] componentsSeparatedByString:@"|"];
		[chords setObject:keys forKey:[temp objectAtIndex:0]];
	}
}

- initWithIn:(NSString *)inFile out:(NSString *)outFile{
	self = [super init];
	lastChord = nil;
	DELTA = 256;
	deltaDivide = 4;
	inData = [[NSFileHandle fileHandleForReadingAtPath:inFile] availableData];
	inLength = [inData length];
	inData = [inData bytes];
	pos = 0;
	self->outFile = outFile;
	track = [[NSMutableData alloc] init];
	contents = [[NSMutableData alloc] init];
	volume = 100;
	
	return self;
}

- (void) writeFile{
	char header[18] = { 0x4D, 0x54, 0x68, 0x64, 0x00, 0x00, 0x00, 0x06,
						0x00, 0x00, 0x00, 0x01, 0x00, 0x40, 0x4D, 0x54,
						0x72, 0x6B };		// MThd header
	char end[3] = { 0xFF, 0x2F, 0x00 };		// MTrk
	char guitar[3] = { 0x00, 0xC0, 0x19 };	// change instrument to guitar
	int length;

	[contents appendBytes:header length:18];
	[self buildTrack];
	length = [track length]+3;
	[contents appendBytes:&length length:4];
	[contents appendBytes:guitar length:3];
	[contents appendData:track];
	[contents appendBytes:end length:3];
	[[NSFileManager defaultManager] createFileAtPath:outFile contents:contents attributes:nil];
}

- (void) buildTrack{
	char c;
	int delta = 0;

	while( pos < inLength ){
		c = inData[pos++];
		if( 'A' <= c && c <= 'G' ){		// a chord
			NSString *chord = [NSString stringWithFormat:@"%c",c];
			int i;
			for( i = 0; i < 13; i++ ){	// start at 7 characters (unicode, 2 bytes)...
				if( pos == inLength )
					break;
				chord = [chord stringByAppendingFormat:@"%c",inData[pos++]];
			}
			while( [chords objectForKey:chord] == nil ){	// find longest match
				chord = [chord substringToIndex:[chord length]-1];
				pos -= 2;		// unicode again
			}
		NSLog( @"%@ at %d", chord, delta );
			[self closeLastChord:delta];
			[self writeChord:chord atDelta:delta];
			delta = DELTA/deltaDivide;
		} else if( c == 'x' || c == 'X' ){	// kill the chord
		NSLog( @"closing %@ at %d", lastChord, delta );
			[self closeLastChord:delta];
			lastChord = nil;
		} else if( c == '-' ){			// a continuation
		NSLog( @"	holding at %d", delta );
			delta += DELTA/deltaDivide;
		} else if( c == '(' ){			// time change
			NSString *num = @"";
			while( (c = inData[pos++]) != ')' )
				num = [num stringByAppendingFormat:@"%c",c];
			deltaDivide = [num intValue];
		} else if( c == '^' ){			// step offset
			NSString *num = @"";
			while( (c = inData[pos++]) != '^' )
				num = [num stringByAppendingFormat:@"%c",c];
			stepOffset = [num intValue];
		} else if( c == ':' ){
			NSString *num = @"";
			while( (c = inData[pos++]) != ':' )
				num = [num stringByAppendingFormat:@"%c",c];
			volume = [num intValue];
		}
	}
	[self closeLastChord:delta];
	[self writeVarTime:0];
}

- (void) writeChord:(NSString *)c atDelta:(int)delta{
	int i;
	NSArray *keys;
NSLog( @"	writeChord" );
	keys = [chords objectForKey:c];
	lastChord = [c retain];
	lastOffset = stepOffset;
	lastVolume = volume;
	[self writeVarTime:delta];
	[self appendNote:[[keys objectAtIndex:0] intValue]+stepOffset state:YES];
	for( i = 1; i < [keys count]; i++ ){
		[self writeVarTime:0];
		[self appendNote:[[keys objectAtIndex:i] intValue]+stepOffset state:YES];
	}
}

- (void) closeLastChord:(int)delta{
	int i;
	NSArray *keys;

	if( lastChord == NULL )
		return;
NSLog( @"	closeChord" );	
	[self writeVarTime:delta];
	[lastChord autorelease];
	keys = [chords objectForKey:lastChord];
	[self appendNote:[[keys objectAtIndex:0] intValue]+lastOffset state:NO];
	for( i = 1; i < [keys count]; i++ ){
		[self writeVarTime:0];
		[self appendNote:[[keys objectAtIndex:i] intValue]+lastOffset state:NO];
	}
}

- (void) appendNote:(int)note state:(BOOL)on{
	char c[3];

	if( on ){
		c[0] = 0x90;
		c[2] = volume;
	} else {
		c[0] = 0x80;
		c[2] = lastVolume;
	}
	c[1] = note;
	
	[track appendBytes:&c length:3];
}

- (void) writeVarTime:(int)value{
	char c[2];
	if( value < 128 ){
		c[0] = value;
		[track appendBytes:&c length:1];
	} else {
		c[0] = value/128 | 0x80;
		c[1] = value % 128;
		[track appendBytes:&c length:2];
	}
}

@end